# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import common


@common.onlyif_mpi
def test_mpi_setup(_mpi_rank, _mpi_size, _mpi_comm):
    rank = _mpi_rank
    size = _mpi_size
    comm = _mpi_comm

    recv = comm.bcast(rank)
    assert recv == 0, "rank={}".format(rank)

    ranks = comm.allgather(rank)
    assert len(ranks) == size
    assert ranks == list(range(size))
